unit main;

interface

uses
  Windows, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ExtCtrls, Grids, StdCtrls, ComCtrls, IniFiles, ImgList,
  ActnList, ToolWin, ActnMan, ActnCtrls, ActnMenus,
  PlatformDefaultStyleActnCtrls, StdActns;

type
  TZMap = record
    Signature: Array [0..3] of AnsiChar;
    TexDefs: Array [0..255] of Array [0..7] of Word;
    TexOrders: Array [0..255] of Array [0..3] of Word;
    CellDefs: Array [0..255] of Byte;
    Levels: Array [0..15] of Array [0..31] of Array[0..31] of Byte;
    LevelEnv: Array [0..15] of Byte;
    AnimData: Array [0..2565] of Byte;
  end;
  TMainForm = class(TForm)
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    ImageList: TImageList;
    MapPopup: TPopupMenu;
    TileList: TImageList;
    TexList1: TImageList;
    TexList2: TImageList;
    TexList3: TImageList;
    StartupTimer: TTimer;
    ActionManager: TActionManager;
    ActionMainMenuBar: TActionMainMenuBar;
    MenuOpenROM: TAction;
    MenuSaveROM: TAction;
    MenuSaveRomAS: TAction;
    MenuExit: TAction;
    MenuUndo: TEditUndo;
    MenuCut: TEditCut;
    MenuCopy: TEditCopy;
    MenuPaste: TEditPaste;
    MenuDelete: TEditDelete;
    MenuTexEdit: TAction;
    MenuAbout: TAction;
    ActnImageList: TImageList;
    ActionToolBar: TActionToolBar;
    PageControl: TPageControl;
    MapTab: TTabSheet;
    ClearButton: TButton;
    FillEdit: TEdit;
    FillButton: TButton;
    ReplaceEdit1: TEdit;
    WithLabel: TLabel;
    ReplaceEdit2: TEdit;
    ReplaceButton: TButton;
    CellDescText: TStaticText;
    CellLabel: TLabel;
    LevelSelectLabel: TLabel;
    LevelSelect: TComboBox;
    TexViewWindowMenu: TAction;
    LegendWindowMenu: TAction;
    MenuPalEdit: TAction;
    procedure FreePopup;
    procedure ReadIni;
    procedure MenuDefClick(Sender: TObject);
    procedure MenuExitClick(Sender: TObject);
    procedure MenuOpenROMClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    function LoadROM: Boolean;
    procedure SaveROM;
    function ConfirmROMSave: Integer;
    function LevelChanged: Boolean;
    procedure LoadLevel(EpisodeNum, LevelNum: Byte);
    procedure SaveLevel;
    function ConfirmLevelSave: Integer;
    procedure LevelSelectChange(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure MenuSaveROMClick(Sender: TObject);
    procedure MenuSaveROMAsClick(Sender: TObject);
    procedure MenuAboutClick(Sender: TObject);
    procedure CalcStats;
    procedure ClearButtonClick(Sender: TObject);
    procedure FillButtonClick(Sender: TObject);
    procedure ReplaceButtonClick(Sender: TObject);
    procedure MenuTexEditClick(Sender: TObject);
    //procedure BuildTextures;
    function HexToInt(HexStr : string) : Int64;
    procedure StartupTimerTimer(Sender: TObject);
    procedure UpdateTexView(Value: Byte);
    procedure LegendWindowMenuExecute(Sender: TObject);
    procedure TexViewWindowMenuExecute(Sender: TObject);
    procedure MenuPalEditExecute(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    var
      TexTilesOffset,E1TexOffset,E2TexOffset,E3TexOffset: Integer;
      ZMap1, ZMap2, ZMap3, CE: TZMap;
      CurPal: Array [0..15] of TColor;
      DataChanged: Boolean;
      CurEp, CurLev: Byte;
  end;

type
  TLevelMap = array of array of Byte;     // Level map is a 32x32 grid
  TBeyondLevel = record
    Offset: LongWord;
    X,Y: Byte
  end;
  TGenesisColor = array [0..1] of Byte;
  TGenesisPal = array [0..15] of TGenesisColor;
  TLevelShortName = array [0..15] of AnsiChar; // Short level name is 16
                                               // characters wdde
  TCellDef = record
    CellType: Byte;
    Name, Description: string
  end;

const
  IniName='ztedit.ini';
  ROMSection='ROM';
  ColorsSection='Colors';
  CellDefsSection='CellDefs';
  BCellDefsSection='BCellDefs';
  ValidIDKey='ValidID';
  ValidID2Key='ValidID2';
  IDOffsetKey='IDOffset';
  ShortNamesOffsetKey='ShortNamesOffset';
  TexTilesOffsetKey='TexTilesOffset';
  PalOffsetKey='PalOffset';
  E1OffsetKey='E1Offset';
  E2OffsetKey='E2Offset';
  E3OffsetKey='E3Offset';
  E1TexOffsetKey='E1TexOffset';
  E2TexOffsetKey='E2TexOffset';
  E3TexOffsetKey='E3TexOffset';
  BE1OffsetKey='BE1Offset';
  BE2OffsetKey='BE2Offset';
  BE3OffsetKey='BE3Offset';
  BeyondE1Section='BE1L';
  BeyondE2Section='BE2L';
  BeyondE3Section='BE3L';
  OffsetKey='Offset';
  XKey='X';
  YKey='Y';
  CellKey='Cell';
  /// defaults ///
  DefaultShortNamesOffset = $30B2;   // Short level names offset (used in start menu)
  DefaultIDOffset = $180;            // ROM cart serial number offset
  DefaultValidID = 'GM T-119146-01'; // Game ROM serial number
  DefaultValidID2 = 'GM T-119146-00';// Game ROM serial number
  DefaultTexTilesOffset = $12EF26;   // Texture tiles offset
  DefaultPalOffset = $21F2;          // Palette offset
  DefaultE1Offset = $15BA0A;         // Episode 1 maps offset
  DefaultE2Offset = $161D24;         // Episode 2 maps offset
  DefaultE3Offset = $16792C;         // Episode 3 maps offset
  DefaultE1TexOffset = $15A10A;      // Episode 1 textures offset
  DefaultE2TexOffset = $160424;      // Episode 2 textures offset
  DefaultE3TexOffset = $16602C;      // Episode 3 textures offset
  DefaultBE1Offset = $AC630;  // Beyond Episode 1 definitions offset
  DefaultBE2Offset = $B0CF6;  // Beyond Episode 1 definitions offset
  DefaultBE3Offset = $B5CBE;  // Beyond Episode 1 definitions offset
  WindowTitle = '%s - %s';
  StatFmt = 'Enemies: %d   Weapons: %d    Items: %d    Sprites: %d';
  CellTypes: Array [0..16] of string = (
             'Empty', 'Wall', 'Upper-right corner', 'Lower-right corner',
             'Lower-left corner', 'Upper-left corner', 'Horizontal door',
             'Vertical door', 'Special', 'Player start', 'Enemy', 'Unknown',
             'Weapon','Item','Special wall', 'Shootable wall', 'Sprite');
  {CurPal: Array [0..15] of TColor = (
  $00E000E0, $00402000, $00806040, $00c0a080,
  $00e0c0c0, $00002020, $00004040, $00e06020,
  $00000040, $00002060, $004080e0, $00204080,
  $00a0e0e0, $000020c0, $00208040, $00000000);}

var
  MainForm: TMainForm;
  Ini: TIniFile;
  IDOffset,ShortNamesOffset,PalOffset,E1Offset,E2Offset,E3Offset,BE1Offset,
  BE2Offset,BE3Offset,BeyondOffset,BackgroundColor,
  BackgroundText,MainColor1,MainColor2,MainColor3,MainText,SpecialColor,
  SpecialText,StartColor,StartText,EnemyColor,EnemyText,UnknownColor,UnknownText,
  WeaponColor,WeaponText,ItemColor,ItemText,SpecialWallColor,SpecialWallText,
  ReservedColor,ReservedText: LongWord;
  ValidID, ValidID2: String;
  CellDefs,BCellDefs: array [0..255] of TCellDef;
  ROM: TMemoryStream; // The game ROM will be loaded in here
  ID: Array [0..13] of AnsiChar; // ROM serial variable
  ShortNames: Array [0..43] of TLevelShortName; // There are 44 records
  CurrentMap: Array of array of Byte; // The map being edited is loaded in here
  BeyondROM: Boolean;
  MapTile: TBitmap;
  ParentPopupItems: array [0..16] of TMenuItem;
  PopupItems: array [0..255] of TMenuItem;
  PPopup: Boolean;
  CurCol,CurRow: Integer;
  BeyondLevels: array [0..23] of TBeyondLevel;
  RomPal: TGenesisPal;

implementation

uses about, TexEdit, Legend, map, texload, texview, PalEdit;

{$R *.dfm}

function TMainForm.HexToInt(HexStr : string) : Int64;
var RetVar : Int64;
    i : byte;
begin
  HexStr := UpperCase(HexStr);
  if HexStr='' then
    HexStr:='00';
  if HexStr[length(HexStr)] = 'H' then
     Delete(HexStr,length(HexStr),1);
  RetVar := 0;

  for i := 1 to length(HexStr) do begin
      RetVar := RetVar shl 4;
      if HexStr[i] in ['0'..'9'] then
         RetVar := RetVar + (byte(HexStr[i]) - 48)
      else
         if HexStr[i] in ['A'..'F'] then
            RetVar := RetVar + (byte(HexStr[i]) - 55)
         else begin
            Retvar := 0;
            break;
         end;
  end;
  Result := RetVar;
end;

function Split(Cmd: String): TCellDef;
var
  S: String;
begin
    S:=Copy(Cmd,1,Pos(';',Cmd)-1);
    Delete(Cmd,1,Pos(';',Cmd));
    Result.CellType:=StrToInt(S);
    S:=Copy(Cmd,1,Pos(';',Cmd)-1);
    Delete(Cmd,1,Pos(';',Cmd));
    Result.Name:=S;
    Result.Description:=Cmd
end;

{procedure TMainForm.BuildTextures;
const
  StatText='Episode %d texture %d of %d';
var
  CurTile,Tile1,Tile2,Tile3,Tile4,Tile5,Tile6,Tile7,Tile8: TBitmap;
  TileBuf: Array [0..511] of Byte;
  Tile: Array [0..31,0..31] of Byte;
  function GetClr1(B: Byte): Byte;
  asm
    shr b,4
    mov result,b
  end;

  function GetClr2(B: Byte): Byte;
  asm
    shl b,4
    shr b,4
    mov result,b
  end;
  procedure BuildTiles;
  var
    I,J,K,L: Word;
  begin
    ROM.Seek(TexTilesOffset,soFromBeginning);
    TileList.Clear;
    for L:= 0 to 246 do
    begin
      K:=0;
      ROM.ReadBuffer(TileBuf,SizeOf(TileBuf));
      CurTile:=TBitmap.Create;
      CurTile.Width:=32;
      CurTile.Height:=32;
      for I := 0 to 15 do
      begin
        for J := 0 to 31 do
        begin
          Tile[I*2,J]:=GetClr1(TileBuf[K]);
          CurTile.Canvas.Pixels[I*2,J]:=CurPal[GetClr1(TileBuf[K])];
          Tile[I*2+1,J]:=GetClr2(TileBuf[K]);
          CurTile.Canvas.Pixels[I*2+1,J]:=CurPal[GetClr2(TileBuf[K])];
          Inc(K);
        end
      end;
      TileList.Add(CurTile,nil);
      CurTile.Free;
    end;
  end;
  procedure BuildTextureEp(Ep: TZMap; EpNum: Byte; var DestList: TImageList);
  var
    I,J,K,L: Word;
  begin
    DestList.Clear;
    for L := 0 to 255 do
    begin
      CurTile:=TBitmap.Create;
      CurTile.Width:=128;
      CurTile.Height:=64;
      Tile1:=TBitmap.Create;
      Tile2:=TBitmap.Create;
      Tile3:=TBitmap.Create;
      Tile4:=TBitmap.Create;
      Tile5:=TBitmap.Create;
      Tile6:=TBitmap.Create;
      Tile7:=TBitmap.Create;
      Tile8:=TBitmap.Create;
      TileList.GetBitmap(Hi(Ep.TexDefs[L,0]),Tile1);
      TileList.GetBitmap(Hi(Ep.TexDefs[L,1]),Tile2);
      TileList.GetBitmap(Hi(Ep.TexDefs[L,2]),Tile3);
      TileList.GetBitmap(Hi(Ep.TexDefs[L,3]),Tile4);
      TileList.GetBitmap(Hi(Ep.TexDefs[L,4]),Tile5);
      TileList.GetBitmap(Hi(Ep.TexDefs[L,5]),Tile6);
      TileList.GetBitmap(Hi(Ep.TexDefs[L,6]),Tile7);
      TileList.GetBitmap(Hi(Ep.TexDefs[L,7]),Tile8);
      for I := 0 to 31 do
        for J := 0 to 31 do
        begin
          CurTile.Canvas.Draw(0,0,Tile1);
          CurTile.Canvas.Draw(0,32,Tile2);
          CurTile.Canvas.Draw(32,0,Tile3);
          CurTile.Canvas.Draw(32,32,Tile4);
          CurTile.Canvas.Draw(64,0,Tile5);
          CurTile.Canvas.Draw(64,32,Tile6);
          CurTile.Canvas.Draw(96,0,Tile7);
          CurTile.Canvas.Draw(96,32,Tile8);
        end;
      Tile1.Free;
      Tile2.Free;
      Tile3.Free;
      Tile4.Free;
      Tile5.Free;
      Tile6.Free;
      Tile7.Free;
      Tile8.Free;
      DestList.Add(CurTile,nil);
      CurTile.Free;
      TexLoadForm.BuildStatLabel.Caption:=Format(StatText,[EpNum,L+1,256]);
      TexLoadForm.BuildProgress.StepIt;
      Application.ProcessMessages;
    end;
  end;
begin
  MainForm.Enabled:=False;
  TexLoadForm.Tag:=1;
  TexLoadForm.Show;
  BuildTiles;
  TexLoadForm.BuildProgress.Position:=0;
  BuildTextureEp(ZMap1,1,TexList1);
  BuildTextureEp(ZMap2,2,TexList2);
  BuildTextureEp(ZMap3,3,TexList3);
  TexLoadForm.Tag:=0;
  TexLoadForm.Close;
  MainForm.Enabled:=True;
  MainForm.BringToFront
end;}

procedure TMainForm.UpdateTexView;
var
  TmpTex: TBitmap;
  TmpList: TImageList;
  X,Y: Byte;
begin
  with TexViewForm do
  begin
    NorthTexture.Picture:=nil;
    SouthTexture.Picture:=nil;
    EastTexture.Picture:=nil;
    WestTexture.Picture:=nil;
    case CurEp of
      1: TmpList:=TexList1;
      2: TmpList:=TexList2;
      3: TmpList:=TexList3;
    end;
    TmpTex:=TBitmap.Create;
    TmpTex.Width:=128;
    TmpTex.Height:=64;
    TmpList.GetBitmap(Hi(CE.TexOrders[Value,0]),NorthTexture.Picture.Bitmap);
    TmpList.GetBitmap(Hi(CE.TexOrders[Value,1]),WestTexture.Picture.Bitmap);
    TmpList.GetBitmap(Hi(CE.TexOrders[Value,2]),EastTexture.Picture.Bitmap);
    TmpList.GetBitmap(Hi(CE.TexOrders[Value,3]),SouthTexture.Picture.Bitmap);
    TmpTex.Free;
    Application.ProcessMessages
  end;
end;

procedure TMainForm.ReadIni;
var
  I: Byte;
begin
  // reading ROM-related variables //
  ValidID:=Ini.ReadString(ROMSection,ValidIDKey,DefaultValidID);
  ValidID2:=Ini.ReadString(ROMSection,ValidID2Key,DefaultValidID2);
  IDOffset:=Ini.ReadInteger(ROMSection,IDOffsetKey,DefaultIDOffset);
  ShortNamesOffset:=Ini.ReadInteger(ROMSection,ShortNamesOffsetKey,
                                    DefaultShortNamesOffset);
  TexTilesOffset:=Ini.ReadInteger(ROMSection,TexTilesOffsetKey,DefaultTexTilesOffset);
  E1Offset:=Ini.ReadInteger(ROMSection,E1OffsetKey,DefaultE1Offset);
  E2Offset:=Ini.ReadInteger(ROMSection,E2OffsetKey,DefaultE2Offset);
  E3Offset:=Ini.ReadInteger(ROMSection,E3OffsetKey,DefaultE3Offset);
  PalOffset:=Ini.ReadInteger(ROMSection,PalOffsetKey,DefaultPalOffset);
  E1TexOffset:=Ini.ReadInteger(ROMSection,E1TexOffsetKey,DefaultE1TexOffset);
  E2TexOffset:=Ini.ReadInteger(ROMSection,E2TexOffsetKey,DefaultE2TexOffset);
  E3TexOffset:=Ini.ReadInteger(ROMSection,E3TexOffsetKey,DefaultE3TexOffset);
  BE1Offset:=Ini.ReadInteger(ROMSection,BE1OffsetKey,DefaultBE1Offset);
  BE2Offset:=Ini.ReadInteger(ROMSection,BE2OffsetKey,DefaultBE2Offset);
  BE3Offset:=Ini.ReadInteger(ROMSection,BE3OffsetKey,DefaultBE3Offset);
  for I := 0 to 7 do
  begin
    BeyondLevels[I].Offset:=Ini.ReadInteger(BeyondE1Section+IntToStr(I+1),OffsetKey,0);
    BeyondLevels[I].X:=Ini.ReadInteger(BeyondE1Section+IntToStr(I+1),XKey,0);
    BeyondLevels[I].Y:=Ini.ReadInteger(BeyondE1Section+IntToStr(I+1),YKey,0)
  end;
    for I := 8 to 13 do
  begin
    BeyondLevels[I].Offset:=Ini.ReadInteger(BeyondE2Section+IntToStr(I-7),OffsetKey,0);
    BeyondLevels[I].X:=Ini.ReadInteger(BeyondE2Section+IntToStr(I-7),XKey,0);
    BeyondLevels[I].Y:=Ini.ReadInteger(BeyondE2Section+IntToStr(I-7),YKey,0)
  end;
    for I := 14 to 23 do
  begin
    BeyondLevels[I].Offset:=Ini.ReadInteger(BeyondE3Section+IntToStr(I-13),OffsetKey,0);
    BeyondLevels[I].X:=Ini.ReadInteger(BeyondE3Section+IntToStr(I-13),XKey,0);
    BeyondLevels[I].Y:=Ini.ReadInteger(BeyondE3Section+IntToStr(I-13),YKey,0)
  end;
  // reading cell definitions //
  for I := 0 to 255 do
  begin
    CellDefs[I]:=Split(Ini.ReadString(CellDefsSection,CellKey+IntToStr(I),'11;Undefined;'));
    BCellDefs[I]:=Split(Ini.ReadString(BCellDefsSection,CellKey+IntToStr(I),'11;empty'));
    if BCellDefs[I].Description='empty' then
      BCellDefs[I]:=CellDefs[I];
  end;
end;

procedure TMainForm.ReplaceButtonClick(Sender: TObject);
var
  I, J: Byte;
  Count: Word;
begin
  Count:=0;
  for I := 0 to MapForm.Map.ColCount do
    for J := 0 to MapForm.Map.RowCount do
      if Uppercase(MapForm.Map.Cells[I,J])=Uppercase(ReplaceEdit1.Text) then
      begin
        MapForm.Map.Cells[I,J]:=Uppercase(ReplaceEdit2.Text);
        Count:=Count+1
      end;
  Application.MessageBox(PWideChar(IntToStr(Count)+' replacements made'),'Information',
                         MB_ICONINFORMATION or MB_OK);
end;

procedure TMainForm.FreePopup;
var
  I: Byte;
begin
  for I:=255 downto 0 do
    PopupItems[I].Free;
  for I := 16 downto 0 do
    ParentPopupItems[I].Free;
  PPopup:=False
end;

procedure TMainForm.MenuDefClick(Sender: TObject);
begin
  with Sender as TMenuItem do
    MapForm.Map.Cells[CurCol,CurRow]:=IntToHex(Tag,2)
end;

procedure TMainForm.LevelSelectChange(Sender: TObject);
var
  LN,EN: Byte;
begin
  MainForm.Enabled:=False;
  if LevelChanged then
    case ConfirmLevelSave of
      mrCancel: Exit;
      mrYes: SaveLevel
    end;
  with MapForm do
  case LevelSelect.ItemIndex of
    0..15: begin LN:=LevelSelect.ItemIndex; EN:=1 end;
    16..31: begin LN:=LevelSelect.ItemIndex-16; EN:=2 end;
    32..47: begin LN:=LevelSelect.ItemIndex-32; EN:=3 end;
  end;
  LoadLevel(EN,LN);
  MainForm.Enabled:=True;
  LevelSelect.SetFocus
end;

function TMainForm.LoadROM: Boolean;
var
  I: Byte;
  J: Integer;
  procedure LoadPal;
  var
    CurClr,CurClr2: String;
    I, J: Byte;
  begin
    ROM.Seek(PalOffset,soFromBeginning);
    ROM.ReadBuffer(RomPal,SizeOf(RomPal));
    for I := 0 to 15 do
    begin
      CurClr2:='$';
      CurClr:=IntToHex(RomPal[I][0],2)+IntToHex(RomPal[I][1],2);
      for J := 1 to Length(CurClr) do
        CurClr2:=CurClr2+CurClr[J]+'0';
      if I<>0 then
        CurPal[I]:=StrToInt(CurClr2)
      else
        CurPal[I]:=StrToInt('$00E000E0')
    end;
  end;
begin
  Result:=False;
  if DataChanged then
  case ConfirmROMSave of
    mrCancel: Exit;
    mrYes: SaveROM
  end;
  DataChanged:=False;
  if not OpenDialog.Execute then
    Exit;
  MainForm.Enabled:=False;
  LevelSelect.Clear;
  ROM.LoadFromFile(OpenDialog.FileName);
  ROM.Seek(IDOffset,soFromBeginning);
  ROM.ReadBuffer(ID,SizeOf(ID)); // Read ROM serial
  if (ID<>ValidID) and (ID<>ValidID2) then // The file is invalid, we won' load it
  begin
    Application.MessageBox('This is not a valid Zero Tolerance ROM file.'#13#10+
                           'A good dump in .BIN (not .SMD) format is required.',
                           'Error',MB_ICONERROR or MB_OK);
    Exit
  end;
  if ID=ValidID2 then
  begin
    Application.MessageBox('Sorry, Beyond Zero Tolerance support has been temporarily '+
                           'removed from this release. It will be back soon.',
                           'Error',MB_ICONERROR or MB_OK);
    Exit
  end;
  BeyondROM:=False;
  if ID=ValidID2 then
    BeyondROM:=True;
  SaveDialog.FileName:=OpenDialog.FileName;
  if not BeyondROM then
  begin
    LoadPal;
    ROM.Seek(ShortNamesOffset,soFromBeginning);
    ROM.ReadBuffer(ShortNames,SizeOf(ShortNames));
    for I := 0 to 43 do
      LevelSelect.Items.Add(ShortNames[I]);
    for I := 44 to 47 do
      LevelSelect.Items.Add('<UNTITLED '+IntToStr(I)+'>');
    ROM.Seek(E1Offset,soFromBeginning);
    ROM.ReadBuffer(ZMap1,SizeOf(ZMap1));
    ROM.Seek(E2Offset,soFromBeginning);
    ROM.ReadBuffer(ZMap2,SizeOf(ZMap2));
    ROM.Seek(E3Offset,soFromBeginning);
    ROM.ReadBuffer(ZMap3,SizeOf(ZMap3));
  end
  else
  begin
    ROM.Seek(BE1Offset,soFromBeginning);
    ROM.ReadBuffer(ZMap1,SizeOf(ZMap1));
    ROM.Seek(BE2Offset,soFromBeginning);
    ROM.ReadBuffer(ZMap2,SizeOf(ZMap2));
    ROM.Seek(BE3Offset,soFromBeginning);
    ROM.ReadBuffer(ZMap3,SizeOf(ZMap3));
    for I := 0 to 7 do
      LevelSelect.Items.Add(BeyondE1Section+IntToStr(I+1));
    for I := 8 to 13 do
      LevelSelect.Items.Add(BeyondE2Section+IntToStr(I-7));
    for I := 14 to 23 do
      LevelSelect.Items.Add(BeyondE3Section+IntToStr(I-13));
  end;
  LevelSelect.ItemIndex:=0;
  for I := 0 to MapForm.Map.ColCount - 1 do
    for J := 0 to MapForm.Map.RowCount - 1 do
      MapForm.Map.Cells[I,J]:='';
  //BuildTextures;
  LoadLevel(1,0);
  MenuSaveROM.Enabled:=True;
  MenuSaveROMAs.Enabled:=True;
  MainForm.Enabled:=True;
  Result:=True;
  CalcStats
end;

procedure TMainForm.SaveROM;
  procedure SavePal;
  var
    SCurrentClr: String;
    CurrentClr: Array [0..1] of Byte;
    I: Byte;
  begin
    ROM.Seek(PalOffset,soFromBeginning);
    for I := 0 to 15 do
    begin
      SCurrentClr:=IntToHex(CurPal[I],8);
      SCurrentClr:=SCurrentClr[1]+SCurrentClr[3]+SCurrentClr[5]+SCurrentClr[7];
      if I=0 then
      begin
        CurrentClr[0]:=0;
        CurrentClr[1]:=0;
      end else
      begin
//        ShowMessage('$'+SCurrentClr[1]+SCurrentClr[2]+', '+'$'+SCurrentClr[3]+SCurrentClr[4]);
        CurrentClr[0]:=StrToInt('$'+SCurrentClr[1]+SCurrentClr[2]);
        CurrentClr[1]:=StrToInt('$'+SCurrentClr[3]+SCurrentClr[4])
      end;
      RomPal[I][0]:=CurrentClr[0];
      RomPal[I][1]:=Currentclr[1];
    end;
    ROM.WriteBuffer(RomPal,SizeOf(RomPal))
  end;
begin
  SavePal;
  ROM.SaveToFile(SaveDialog.FileName);
  DataChanged:=False
end;

procedure TMainForm.StartupTimerTimer(Sender: TObject);
begin
  StartupTimer.Enabled:=False;
  if not LoadROM then
    Application.Terminate;
end;

procedure TMainForm.TexViewWindowMenuExecute(Sender: TObject);
begin
  if TexViewWindowMenu.Checked then
    TexViewForm.Destroy
  else
    Application.CreateForm(TTexViewForm, TexViewForm)
end;

function TMainForm.ConfirmROMSave: Integer;
begin
  Result:=Application.MessageBox('ROM data has been changed. Do you wish to '+
                                 'save changes?','Question',MB_ICONQUESTION or
                                 MB_YESNOCANCEL)
end;

procedure TMainForm.LoadLevel;
var
  I, J: Byte;
  CI: string;
begin
  case EpisodeNum of
    1: CE:=ZMap1;
    2: CE:=ZMap2;
    3: CE:=ZMap3;
  end;
  CellLabel.Caption:='';
  CellDescText.Caption:='';
  if PPopup then
    FreePopup;
  MapForm.Map.ColCount:=32;
  MapForm.Map.RowCount:=32;
  if (EpisodeNum=3) and (LevelNum>11) then
    MapForm.Caption:='?'
  else
    MapForm.Caption:='Map - '+ShortNames[LevelNum+16*(EpisodeNum-1)];// Show level name below the map grid}
  for I := 0 to 16 do
  begin
    ParentPopupItems[I]:=TMenuItem.Create(MapPopup);
    MapPopup.Items.Add(ParentPopupItems[I]);
    ParentPopupItems[I].Name:='PopupParent'+IntToStr(I);
    ParentPopupItems[I].Caption:=CellTypes[I]
  end;
  for I := 0 to 255 do
  begin
    CI:=IntToHex(I,2);
    PopupItems[I]:=TMenuItem.Create(MapPopup);
    MapPopup.Items[CellDefs[CE.CellDefs[I]].CellType].Add(PopupItems[I]);
    PopupItems[I].Name:='MenuDef'+IntToStr(I);
    PopupItems[I].Caption:='['+CI+']: '+CellDefs[CE.CellDefs[I]].Name;
    PopupItems[I].Tag:=I;
    PopupItems[I].OnClick:=MenuDefClick
  end;
  PPopup:=True;
  SetLength(CurrentMap,32,32);
  for I := 0 to 31 do
    for J := 0 to 31 do
    begin
      CurrentMap[I,J]:=CE.Levels[LevelNum,I,J];
      MapForm.Map.Cells[J,I]:=IntToHex(CurrentMap[I,J],2)
    end;
  CurEp:=EpisodeNum;
  CurLev:=LevelNum;
  if not MapForm.Visible then
    MapForm.Show;
  MapForm.Map.Refresh; // Redraw the grid
  CalcStats
end;

procedure TMainForm.LegendWindowMenuExecute(Sender: TObject);
begin
  if LegendWindowMenu.Checked then
    LegendForm.Destroy
  else
    Application.CreateForm(TLegendForm, LegendForm)
end;

function TMainForm.LevelChanged: Boolean;
var
  I,J: Byte;
  NewMap: TLevelMap;
begin
  Result:=False;
  SetLength(NewMap,MapForm.Map.RowCount,MapForm.Map.ColCount);
  for I := 0 to MapForm.Map.RowCount-1 do
    for J := 0 to MapForm.Map.ColCount-1 do
    begin
      NewMap[I,J]:=HexToInt(MapForm.Map.Cells[J,I]);
      if NewMap[I,J]<>CurrentMap[I,J] then
        Result:=True
    end
end;

function TMainForm.ConfirmLevelSave: Integer;
begin
  Result:=Application.MessageBox('Level data has been changed. Do you wish to '+
                                 'save changes?','Question',MB_ICONQUESTION or
                                 MB_YESNOCANCEL)
end;

procedure TMainForm.MenuSaveROMAsClick(Sender: TObject);
begin
  if LevelChanged then
    case ConfirmLevelSave of
      mrYes: SaveLevel;
      mrCancel: Exit
    end;
  if SaveDialog.Execute then
    SaveROM
end;

procedure TMainForm.SaveLevel;
var
  NewMap: Array of Array of Byte;
  I,J: Byte;
begin
  case CurEp of
    1: begin ROM.Seek(E1Offset,soFromBeginning); CE:=ZMap1; end;// Ep. 1 maps
    2: begin ROM.Seek(E2Offset,soFromBeginning); CE:=ZMap2; end;// Ep. 2 maps
    3: begin ROM.Seek(E3Offset,soFromBeginning); CE:=ZMap3; end;// Ep. 3 maps
  else
    Exit; // Level # is out of bounds so we're not saving it
  end;
  SetLength(NewMap,MapForm.Map.RowCount,MapForm.Map.ColCount);
  for I := 0 to MapForm.Map.RowCount-1 do
    for J := 0 to MapForm.Map.ColCount-1 do
    begin
      NewMap[I,J]:=HexToInt(MapForm.Map.Cells[J,I]);
      CurrentMap[I,J]:=NewMap[I,J];
      CE.Levels[CurLev,I,J]:=CurrentMap[I,J];
    end;
  case CurEp of
    1: ZMap1:=CE;
    2: ZMap2:=CE;
    3: ZMap3:=CE;
  end;
  ROM.WriteBuffer(CE,SizeOf(CE));
  DataChanged:=True
end;

procedure TMainForm.FillButtonClick(Sender: TObject);
var
  I, J: Byte;
begin
  if Application.MessageBox('Are you sure you want to fill this map?','Question',
                            MB_ICONQUESTION or MB_YESNO)=mrYes then
  for I := 0 to MapForm.Map.ColCount do
    for J := 0 to MapForm.Map.RowCount do
      MapForm.Map.Cells[I,J]:=FillEdit.Text
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if LevelChanged then
    case ConfirmLevelSave of
      mrCancel: begin Action:=caNone; Exit end;
      mrYes: SaveLevel
    end;
  if DataChanged then
    case ConfirmROMSave of
      mrCancel: begin Action:=caNone; Exit end;
      mrYes: SaveROM
    end
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  PPopup:=False;
  Ini:=TIniFile.Create(ExtractFilePath(Application.ExeName)+IniName);
  ReadIni;
  DataChanged:=False;
  ROM:=TMemoryStream.Create; // Initialize ROM content storage
  MapTile:=TBitmap.Create
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FreePopup;
  MapTile.Free;
  Ini.Free;
  ROM.Free // Destroy ROM content storage
end;

procedure TMainForm.ClearButtonClick(Sender: TObject);
var
  I, J: Byte;
begin
  if Application.MessageBox('Are you sure you want to clear this map?','Question',
                            MB_ICONQUESTION or MB_YESNO)=mrYes then
  for I := 0 to MapForm.Map.ColCount do
    for J := 0 to MapForm.Map.RowCount do
      MapForm.Map.Cells[I,J]:='00'
end;

procedure TMainForm.CalcStats;
var
  Enemies, Weapons, Items, Sprites: Word;
  I, J: Byte;
begin
  Enemies:=0;
  Weapons:=0;
  Items:=0;
  Sprites:=0;
  for I := 0 to MapForm.Map.ColCount do
    for J := 0 to MapForm.Map.RowCount do
      if BeyondROM then
      case BCellDefs[CE.CellDefs[HexToInt(MapForm.Map.Cells[I,J])]].CellType of
        10: Enemies:=Enemies+1;
        12: Weapons:=Weapons+1;
        13: Items:=Items+1;
        16: Sprites:=Sprites+1
      end else
      case CellDefs[CE.CellDefs[HexToInt(MapForm.Map.Cells[I,J])]].CellType of
        10: Enemies:=Enemies+1;
        12: Weapons:=Weapons+1;
        13: Items:=Items+1;
        16: Sprites:=Sprites+1
      end;
  MapForm.LevelStatBar.Panels.Items[0].Text:='Enemies: '+IntToStr(Enemies);
  MapForm.LevelStatBar.Panels.Items[1].Text:='Weapons: '+IntToStr(Weapons);
  MapForm.LevelStatBar.Panels.Items[2].Text:='Items: '+IntToStr(Items);
  MapForm.LevelStatBar.Panels.Items[3].Text:='Sprites: '+IntToStr(Sprites);
end;

procedure TMainForm.MenuAboutClick(Sender: TObject);
begin
  AboutForm.ShowModal
end;

procedure TMainForm.MenuExitClick(Sender: TObject);
begin
  Close
end;

procedure TMainForm.MenuOpenROMClick(Sender: TObject);
begin
  LoadROM
end;

procedure TMainForm.MenuPalEditExecute(Sender: TObject);
begin
  PalEditForm.ShowModal;
end;

procedure TMainForm.MenuSaveROMClick(Sender: TObject);
begin
  if LevelChanged then
    case ConfirmLevelSave of
      mrYes: SaveLevel;
      mrCancel: Exit
    end;
  SaveROM
end;

procedure TMainForm.MenuTexEditClick(Sender: TObject);
begin
  Application.CreateForm(TTexEditForm, TexEditForm);
  TexEditForm.ShowModal
end;

end.
