unit map;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, ComCtrls, ExtCtrls, StdCtrls;

type
  TMapForm = class(TForm)
    Map: TStringGrid;
    LevelStatBar: TStatusBar;
    procedure MapDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
      State: TGridDrawState);
    procedure MapContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure MapSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure MapSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: string);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MapForm: TMapForm;

implementation

uses main, texview;

{$R *.dfm}

procedure TMapForm.MapContextPopup(Sender: TObject; MousePos: TPoint;
  var Handled: Boolean);
var
  CanSelect: Boolean;
begin
  Map.MouseToCell(MousePos.X,MousePos.Y,CurCol,CurRow);
  MapSelectCell(Map,CurCol,CurRow,CanSelect)
end;

procedure TMapForm.MapDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  TileNum: Byte;
  CL: Word;
  S: String;
begin
  with MainForm do
  begin
    CL:=Length(Map.Cells[ACol,ARow]);
    if CL<2 then
      Map.Cells[ACol,ARow]:='0'+Map.Cells[ACol,ARow];
    if CL>2 then
    begin
      S:=Map.Cells[ACol,ARow];
      Delete(S,3,CL-1);
      Map.Cells[ACol,ARow]:=S
    end;
    if Map.Cells[ACol,ARow]<>Uppercase(Map.Cells[ACol,ARow]) then
      Map.Cells[ACol,ARow]:=Uppercase(Map.Cells[ACol,ARow]);
    if BeyondROM then
      TileNum:=BCellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].CellType
    else
      TileNum:=CellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].CellType;
    if TileNum in [8,10,13,16] then
      Map.Canvas.Font.Color:=clWhite
    else
      Map.Canvas.Font.Color:=clBlack;
    ImageList.GetBitmap(TileNum,MapTile);
    Map.Canvas.StretchDraw(Rect, MapTile);
    if not (TileNum in [0,6,7]) then
    with Map.Canvas, Rect do
      begin
        Map.Canvas.Brush.Style:=bsClear;
        Map.Canvas.Font.Name:='Small Fonts';
        Map.Canvas.Font.Size:=6;
        case TileNum of
          2: Map.Canvas.TextOut(Left+6, Top, UpperCase(Map.Cells[ACol,ARow]));
          3: Map.Canvas.TextOut(Left+6, Top+7, UpperCase(Map.Cells[ACol,ARow]));
          4: Map.Canvas.TextOut(Left, Top+7, UpperCase(Map.Cells[ACol,ARow]));
          5: Map.Canvas.TextOut(Left, Top, UpperCase(Map.Cells[ACol,ARow]));
        else Map.Canvas.TextOut(Left+3, Top+3, UpperCase(Map.Cells[ACol,ARow]))
        end
      end;
      if gdSelected in State then
      begin
//        Map.Canvas.Pen.Style:=bsSolid;
        Map.Canvas.Pen.Color:=clRed;
        Map.Canvas.Rectangle(Rect.Left,Rect.Top,Rect.Right,Rect.Top+2);
        Map.Canvas.Rectangle(Rect.Left,Rect.Top,Rect.Left+2,Rect.Bottom);
        Map.Canvas.Rectangle(Rect.Left,Rect.Bottom-2,Rect.Right,Rect.Bottom);
        Map.Canvas.Rectangle(Rect.Right-2,Rect.Top,Rect.Right,Rect.Bottom);
      end;
  end;
end;

procedure TMapForm.MapSelectCell(Sender: TObject; ACol, ARow: Integer;
  var CanSelect: Boolean);
const
  CellLabelText='[%s] %s';
var
  CNum, CDef, CName, CDescr: String;
begin
  with MainForm do
  begin
    CNum:=Map.Cells[ACol,ARow];
    if BeyondROM then
    begin
        CDef:=CellTypes[BCellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].CellType];
        CName:=BCellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].Name
    end
    else
    begin
        CDef:=CellTypes[CellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].CellType];
        CName:=CellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].Name
    end;
    if CName<>'' then
      CDef:=CDef+': '+CName;
    if BeyondROM then
      CDescr:=BCellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].Description
    else
      CDescr:=CellDefs[CE.CellDefs[HexToInt(Map.Cells[ACol,ARow])]].Description;
    if CDescr<>'' then
      CellDescText.Caption:=CDescr
    else
      CellDescText.Caption:='';
    CellLabel.Caption:=Format(CellLabelText,[CNum,CDef]);
    try
      UpdateTexView(HexToInt(Map.Cells[ACol,ARow]))
    except
    end;
  end;
end;

procedure TMapForm.MapSetEditText(Sender: TObject; ACol, ARow: Integer;
  const Value: string);
begin
  MainForm.CalcStats;
  MainForm.UpdateTexView(MainForm.HexToInt(Value))
end;

end.
